#!/usr/bin/env python

# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import signal
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from logging import CRITICAL, DEBUG, ERROR, Formatter, getLogger, INFO, NOTSET, StreamHandler, WARNING
from logging.handlers import RotatingFileHandler

from prometheus_client.core import CollectorRegistry
from pysyncobj import SyncObjConf

from cockatrice import NAME, VERSION
from cockatrice.command import add_node, delete_node, get_snapshot, get_status
from cockatrice.index_node import IndexNode


def signal_handler(signal, frame):
    sys.exit(0)


def server_handler(args):
    # create logger and handler
    logger = getLogger(NAME)
    log_handler = StreamHandler()

    # determine log destination
    if args.log_file is not None:
        log_handler = RotatingFileHandler(args.log_file, 'a+', maxBytes=args.log_file_max_bytes,
                                          backupCount=args.log_file_backup_count)

    # determine log level
    if args.log_level in ['CRITICAL', 'FATAL']:
        logger.setLevel(CRITICAL)
        log_handler.setLevel(CRITICAL)
    elif args.log_level == 'ERROR':
        logger.setLevel(ERROR)
        log_handler.setLevel(ERROR)
    elif args.log_level in ['WARNING', 'WARN']:
        logger.setLevel(WARNING)
        log_handler.setLevel(WARNING)
    elif args.log_level == 'INFO':
        logger.setLevel(INFO)
        log_handler.setLevel(INFO)
    elif args.log_level == 'DEBUG':
        logger.setLevel(DEBUG)
        log_handler.setLevel(DEBUG)
    elif args.log_level == 'NOTSET':
        logger.setLevel(NOTSET)
        log_handler.setLevel(NOTSET)
    else:
        logger.setLevel(INFO)
        log_handler.setLevel(INFO)

    # set log format
    handler_format = Formatter('%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)d - %(message)s')
    log_handler.setFormatter(handler_format)

    # add log handler
    logger.addHandler(log_handler)

    # create http logger and handler
    http_logger = getLogger(NAME + '_http')
    http_log_handler = StreamHandler()

    # determine http log destination
    if args.http_log_file is not None:
        http_log_handler = RotatingFileHandler(args.http_log_file, 'a+', maxBytes=args.http_log_file_max_bytes,
                                               backupCount=args.http_log_file_backup_count)

    # determine http log level
    http_logger.setLevel(INFO)
    http_log_handler.setLevel(INFO)

    # set http log format
    http_handler_format = Formatter('%(message)s')
    http_log_handler.setFormatter(http_handler_format)

    # add http log handler
    http_logger.addHandler(http_log_handler)

    # metrics registry
    metrics_registry = CollectorRegistry()

    # sync config
    os.makedirs(os.path.dirname(args.snapshot_file), exist_ok=True)
    conf = SyncObjConf()
    conf.fullDumpFile = args.snapshot_file
    conf.logCompactionMinEntries = args.log_compaction_min_entries
    conf.logCompactionMinTime = args.log_compaction_min_time
    conf.dynamicMembershipChange = True
    conf.validate()

    index_node = None
    try:
        index_node = IndexNode(host=args.host, port=args.port, seed_addr=args.seed_addr, conf=conf,
                               index_dir=args.index_dir, http_port=args.http_port, logger=logger,
                               http_logger=http_logger, metrics_registry=metrics_registry)
        while True:
            signal.pause()
    except Exception as ex:
        print(ex)
    finally:
        if index_node is not None:
            index_node.stop()


def status_handler(args):
    print(json.dumps(get_status(bind_addr=args.bind_addr, timeout=0.5)))


def join_handler(args):
    print(json.dumps(add_node(args.join_addr, bind_addr=args.bind_addr, timeout=0.5)))


def leave_handler(args):
    print(json.dumps(delete_node(args.leave_addr, bind_addr=args.bind_addr, timeout=0.5)))


def snapshot_handler(args):
    snapshot = get_snapshot(bind_addr=args.bind_addr, timeout=0.5)
    if args.output_file is None:
        print(snapshot)
    else:
        with open(args.output_file, 'wb') as f:
            f.write(snapshot)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGQUIT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    parser = ArgumentParser(description='cockatrice command', formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version='cockatrice {0}'.format(VERSION))

    subparsers = parser.add_subparsers()

    # server
    parser_server = subparsers.add_parser('server', help='see `server --help`',
                                          formatter_class=ArgumentDefaultsHelpFormatter)
    parser_server.add_argument('--host', dest='host', default='localhost', metavar='HOST', type=str,
                               help='the host address to listen on for peer traffic')
    parser_server.add_argument('--port', dest='port', default=7070, metavar='PORT', type=int,
                               help='the port to listen on for peer traffic')
    parser_server.add_argument('--seed-addr', dest='seed_addr', default=None, metavar='SEED_ADDR', type=str,
                               help='the address of the node in the existing cluster')
    parser_server.add_argument('--snapshot-file', dest='snapshot_file', default='/tmp/cockatrice/snapshot.zip',
                               metavar='SNAPSHOT_FILE', type=str, help='file to store snapshot of all indices')
    parser_server.add_argument('--log-compaction-min-entries', dest='log_compaction_min_entries', default=5000,
                               metavar='LOG_COMPACTION_MIN_ENTRIES', type=int,
                               help='log-compaction interval min entries')
    parser_server.add_argument('--log-compaction-min-time', dest='log_compaction_min_time', default=300,
                               metavar='LOG_COMPACTION_MIN_TIME', type=int,
                               help='log-compaction interval min time in seconds')
    parser_server.add_argument('--index-dir', dest='index_dir', default='/tmp/cockatrice/index', metavar='INDEX_DIR',
                               type=str, help='index dir')
    parser_server.add_argument('--http-port', dest='http_port', default=8080,
                               metavar='HTTP_PORT', type=int, help='the port to listen on for HTTP traffic')
    parser_server.add_argument('--log-level', dest='log_level', default='DEBUG', metavar='LOG_LEVEL', type=str,
                               help='log level')
    parser_server.add_argument('--log-file', dest='log_file', default=None, metavar='LOG_FILE', type=str,
                               help='log file')
    parser_server.add_argument('--log-file-max-bytes', dest='log_file_max_bytes', default=512000000,
                               metavar='LOG_FILE_MAX_BYTES', type=int, help='log file max bytes')
    parser_server.add_argument('--log-file-backup-count', dest='log_file_backup_count', default=5,
                               metavar='LOG_FILE_BACKUP_COUNT', type=int, help='log file backup count')
    parser_server.add_argument('--http-log-file', dest='http_log_file', default=None, metavar='HTTP_LOG_FILE', type=str,
                               help='http log file')
    parser_server.add_argument('--http-log-file-max-bytes', dest='http_log_file_max_bytes', default=512000000,
                               metavar='HTTP_LOG_FILE_MAX_BYTES', type=int, help='http log file max bytes')
    parser_server.add_argument('--http-log-file-backup-count', dest='http_log_file_backup_count', default=5,
                               metavar='HTTP_LOG_FILE_BACKUP_COUNT', type=int, help='http log file backup count')
    parser_server.set_defaults(handler=server_handler)

    # status
    parser_status = subparsers.add_parser('status', help='see `status --help`',
                                          formatter_class=ArgumentDefaultsHelpFormatter)
    parser_status.add_argument('--bind-addr', dest='bind_addr', default='localhost:7070', metavar='BIND_ADDR', type=str,
                               help='the address to listen on for peer traffic')
    parser_status.set_defaults(handler=status_handler)

    # join
    parser_join = subparsers.add_parser('join', help='see `join --help`', formatter_class=ArgumentDefaultsHelpFormatter)
    parser_join.add_argument('--bind-addr', dest='bind_addr', default='localhost:7070', metavar='BIND_ADDR', type=str,
                             help='the address to listen on for peer traffic')
    parser_join.add_argument('--join-addr', dest='join_addr', default=None, metavar='JOIN_ADDR', type=str,
                             help='the address of node to join to the cluster')
    parser_join.set_defaults(handler=join_handler)

    # leave
    parser_leave = subparsers.add_parser('leave', help='see `leave --help`',
                                         formatter_class=ArgumentDefaultsHelpFormatter)
    parser_leave.add_argument('--bind-addr', dest='bind_addr', default='localhost:7070',
                              metavar='BIND_ADDR', type=str, help='the address to listen on for peer traffic')
    parser_leave.add_argument('--leave-addr', dest='leave_addr', default=None, metavar='LEAVE_ADDR', type=str,
                              help='the address of node to leave from the cluster')
    parser_leave.set_defaults(handler=leave_handler)

    # snapshot
    parser_snapshot = subparsers.add_parser('snapshot', help='see `snapshot --help`',
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser_snapshot.add_argument('--bind-addr', dest='bind_addr', default='localhost:7070',
                                 metavar='BIND_ADDR', type=str, help='the address to listen on for peer traffic')
    parser_snapshot.add_argument('--output-file', dest='output_file', default=None, metavar='OUTPUT_FILE', type=str,
                                 help='the snapshot file destination')
    parser_snapshot.set_defaults(handler=snapshot_handler)

    args = parser.parse_args()
    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from cockatrice.analysis import JanomeTokenizer


class TestJanomeTokenizer(unittest.TestCase):
    def test_janome_tokenizer(self):
        value = u'すもももももももものうち'

        tokenizer = JanomeTokenizer()

        expected = ['すもも', 'も', 'もも', 'も', 'もも', 'の', 'うち']
        actual = [t.text for t in tokenizer(value)]

        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
import re
import threading
import time
import zipfile
from logging import getLogger

import pysyncobj.pickle as pickle
from pysyncobj import replicated, SyncObj, SyncObjConf
from whoosh.filedb.filestore import FileStorage
from whoosh.qparser import QueryParser


class IndexServer(SyncObj):
    def __init__(self, host='localhost', port=7070, peer_addrs=[], conf=SyncObjConf(),
                 index_dir='/tmp/cockatrice/index', logger=getLogger()):
        self.__logger = logger

        self.__lock = threading.RLock()

        self.__bind_addr = '{0}:{1}'.format(host, port)
        self.__peer_addrs = peer_addrs
        self.__index_dir = index_dir
        self.__conf = conf
        self.__conf.serializer = self.__serialize
        self.__conf.deserializer = self.__deserialize
        self.__conf.validate()

        self.__indices = {}

        self.__file_storage = None

        self.__logger.info('starting index server: {0}, {1}'.format(self.__bind_addr, self.__peer_addrs))

        self.__logger.info('creating file storage for indices on {0}'.format(self.__index_dir))
        os.makedirs(self.__index_dir, exist_ok=True)
        self.__file_storage = FileStorage(self.__index_dir, supports_mmap=True, readonly=False, debug=False)

        super(IndexServer, self).__init__(self.__bind_addr, self.__peer_addrs, conf=self.__conf)

        self.__logger.info('index server started')

        # open existing indices on startup
        self.__logger.info('opening existing indices')
        self.__open_existing_indices()
        self.__logger.info('existing indices opened')

        # waiting for the preparation to be completed
        while not self.isReady():
            # recovering data
            self.__logger.info('waiting for the cluster to be ready')
            time.sleep(1)
        self.__logger.info('the index server is ready')

    def stop(self):
        self.__logger.info('stopping index server')
        for index_name in self.__indices.keys():
            self.close_index(index_name)
        self.destroy()
        self.__logger.info('index server stopped')

    def destroy(self):
        super().destroy()

    def __onUtilityMessage(self, conn, message):
        try:
            if message[0] == 'status':
                conn.send(self.getStatus())
                return True
            elif message[0] == 'add':
                self.addNodeToCluster(message[1],
                                      callback=functools.partial(self.__utilityCallback, conn=conn, cmd='ADD',
                                                                 node=message[1]))
                return True
            elif message[0] == 'remove':
                if message[1] == self.__selfNodeAddr:
                    conn.send('FAIL REMOVE ' + message[1])
                else:
                    self.removeNodeFromCluster(message[1], callback=functools.partial(self.__utilityCallback, conn=conn,
                                                                                      cmd='REMOVE', node=message[1]))
                return True
            elif message[0] == 'set_version':
                self.setCodeVersion(message[1],
                                    callback=functools.partial(self.__utilityCallback, conn=conn, cmd='SET_VERSION',
                                                               node=str(message[1])))
                return True
            elif message[0] == 'get_snapshot':
                with open(self.__conf.fullDumpFile, 'rb') as f:
                    conn.send(f.read())
                return True
        except Exception as e:
            conn.send(str(e))
            return True

    def __serialize(self, filename, raft_data):
        self.__logger.info('creating snapshot file: {0}'.format(filename))
        try:
            with self.__lock:
                with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as f:
                    for index_filename in self.__file_storage:
                        abs_index_path = os.path.join(self.__file_storage.folder, index_filename)
                        self.__logger.info('storing index file: {0}'.format(abs_index_path))
                        f.write(abs_index_path, index_filename)
                    self.__logger.info('storing raft transaction log')
                    f.writestr('raft.bin', pickle.dumps(raft_data))
                self.__logger.info('snapshot file created: {0}'.format(filename))
        except Exception as ex:
            self.__logger.error(ex)

    def __deserialize(self, filename):
        self.__logger.info('restoring from snapshot: {0}'.format(self.__conf.fullDumpFile))
        try:
            with self.__lock:
                self.__logger.info('cleaning {0}'.format(self.__file_storage.folder))
                self.__file_storage.destroy()
                self.__file_storage.create()

                with zipfile.ZipFile(filename, 'r') as f:
                    for member in f.namelist():
                        if member != 'raft.bin':
                            self.__logger.info(
                                'restoring {0}'.format(os.path.join(self.__file_storage.folder, member)))
                            f.extract(member, path=self.__file_storage.folder)
                    return pickle.loads(f.read('raft.bin'))
        except Exception as ex:
            self.__logger.error(ex)

    def get_file_storage(self):
        return self.__file_storage

    def __index_exists(self, index_name):
        return self.__file_storage.index_exists(indexname=index_name)

    def __get_existing_index_names(self):
        index_names = []

        pattern_toc = re.compile('^_(.+)_\\d+\\.toc$')
        self.__logger.info('seeking indices on {0}'.format(self.__file_storage.folder))
        for filename in self.__file_storage:
            match = pattern_toc.search(filename)
            if match:
                self.__logger.debug(
                    '{0} found in file storage on {1}'.format(match.group(1), self.__file_storage.folder))
                index_names.append(match.group(1))

        return index_names

    def __open_existing_indices(self):
        for index_name in self.__get_existing_index_names():
            self.open_index(index_name, schema=None, _doApply=False)

    @replicated
    def open_index(self, index_name, schema=None):
        index = None

        try:
            self.__logger.info('opening {0} in file storage on {1}'.format(index_name, self.__file_storage.folder))
            if not self.__index_exists(index_name):
                raise KeyError(
                    '{0} does not exist in file storage on {1}'.format(index_name, self.__file_storage.folder))
            if index_name in self.__indices:
                index = self.__indices[index_name]
                self.__logger.info(
                    '{0} in file storage on {1} already opened'.format(index_name, self.__file_storage.folder))
            else:
                index = self.__file_storage.open_index(indexname=index_name, schema=schema)
                self.__indices[index_name] = index
                self.__logger.info('{0} in file storage on {1} opened'.format(index_name, self.__file_storage.folder))
        except Exception as ex:
            self.__logger.info(
                'failed to open {0} in file storage on {1}: {2}'.format(index_name, self.__file_storage.folder, ex))

        return index

    @replicated
    def close_index(self, index_name):
        try:
            index = self.__indices.pop(index_name)
            self.__logger.info('closing {0} in file storage on {1}'.format(index_name, index.storage.folder))
            if index is None:
                raise KeyError('{0} in file storage on {1} already closed'.format(index_name, index.storage.folder))
            else:
                index.close()
            self.__logger.info('{0} in file storage on {1} closed'.format(index_name, index.storage.folder))
        except Exception as ex:
            self.__logger.error(
                'failed to close {0} in file storage on {1}: {2}'.format(index_name, index.storage.folder, ex))

        return index

    @replicated
    def create_index(self, index_name, schema):
        index = None

        try:
            if self.__index_exists(index_name):
                self.open_index(index_name, schema=schema, _doApply=True)
            else:
                self.__logger.info('creating {0} in file storage on {1}'.format(index_name, self.__file_storage.folder))
                self.__indices[index_name] = self.__file_storage.create_index(schema, indexname=index_name)
                index = self.__indices[index_name]
                self.__logger.info('{0} in file storage on {1} created'.format(index_name, self.__file_storage.folder))
        except Exception as ex:
            self.__logger.error(
                'failed to close {0} in file storage on {1}: {2}'.format(index_name, self.__file_storage.folder, ex))

        return index

    @replicated
    def delete_index(self, index_name):
        success = False

        # close index
        self.close_index(index_name, _doApply=True)

        # delete index files
        try:
            self.__logger.info('deleting {0} in file storage on {1}'.format(index_name, self.__file_storage.folder))
            pattern_toc = re.compile('^_{0}_(\\d+)\\.toc$'.format(index_name))
            for filename in self.__file_storage:
                if re.match(pattern_toc, filename):
                    self.__logger.info('deleting {0}'.format(os.path.join(self.__file_storage.folder, filename)))
                    self.__file_storage.delete_file(filename)
            pattern_seg = re.compile('^{0}_([a-z0-9]+)\\.seg$'.format(index_name))
            for filename in self.__file_storage:
                if re.match(pattern_seg, filename):
                    self.__logger.info('deleting {0}'.format(os.path.join(self.__file_storage.folder, filename)))
                    self.__file_storage.delete_file(filename)
            pattern_lock = re.compile('^{0}_WRITELOCK$'.format(index_name))
            for filename in self.__file_storage:
                if re.match(pattern_lock, filename):
                    self.__logger.info('deleting {0}'.format(os.path.join(self.__file_storage.folder, filename)))
                    self.__file_storage.delete_file(filename)
            success = True
            self.__logger.info('{0} in file storage on {1} deleted'.format(index_name, self.__file_storage.folder))
        except Exception as ex:
            self.__logger.error(
                'failed to delete {0} in file storage on {1}: {2}'.format(index_name, self.__file_storage.folder, ex))

        return success

    def get_indices(self):
        return self.__indices

    def get_index(self, index_name):
        index = self.get_indices().get(index_name)
        if index is None:
            msg = '{0} is not available'.format(index_name)
            self.__logger.error(msg)
            raise KeyError(msg)
        return index

    @replicated
    def optimize_index(self, index_name):
        success = False

        try:
            index = self.get_index(index_name)
            self.__logger.info('optimizing {0} in file storage on {1}'.format(index_name, index.storage.folder))
            index.optimize()
            success = True
            self.__logger.info('{0} in file storage on {1} optimized'.format(index_name, index.storage.folder))
        except Exception as ex:
            self.__logger.error('failed to optimize {0}: {1}'.format(index_name, ex))

        return success

    def get_doc_count(self, index_name):
        try:
            cnt = self.get_index(index_name).doc_count()
        except Exception as ex:
            self.__logger.error('failed to get document count in {0}'.format(index_name))
            raise ex

        return cnt

    def get_writer(self, index_name):
        try:
            writer = self.get_index(index_name).writer()
        except Exception as ex:
            self.__logger.error('failed to get index writer in {0}'.format(index_name))
            raise ex

        return writer

    def get_schema(self, index_name):
        try:
            schema = self.get_index(index_name).schema
        except Exception as ex:
            self.__logger.error('failed to get index schema in {0}'.format(index_name))
            raise ex

        return schema

    def get_searcher(self, index_name, weighting=None):
        try:
            if weighting is None:
                searcher = self.get_index(index_name).searcher()
            else:
                searcher = self.get_index(index_name).searcher(weighting=weighting)
        except Exception as ex:
            self.__logger.error('failed to get index searcher in {0}'.format(index_name))
            raise ex

        return searcher

    @replicated
    def put_document(self, index_name, doc_id, fields):
        success = False

        try:
            self.__logger.debug('putting document in {0}: {1} {2}'.format(index_name, doc_id, fields))

            writer = self.get_writer(index_name)
            doc = {
                self.get_schema(index_name).get_unique_field(): doc_id
            }
            doc.update(fields)
            try:
                writer.update_document(**doc)
                success = True
            except Exception as ex:
                self.__logger.error('failed to put document in {0}: {1}'.format(index_name, doc))
                raise ex
            finally:
                if success:
                    writer.commit()
                else:
                    writer.cancel()
        except Exception as ex:
            self.__logger.error('failed to index document in {0}: {1}'.format(index_name, ex))

        return success

    @replicated
    def delete_document(self, index_name, doc_id):
        success = False

        try:
            self.__logger.debug('deleting document in {0}: {1}'.format(index_name, doc_id))

            unique_field = self.get_schema(index_name).get_unique_field()
            writer = self.get_writer(index_name)
            try:
                writer.delete_by_term(unique_field, doc_id)
                success = True
            except Exception as ex:
                self.__logger.error('failed to delete document in {0}: {1}:{2}'.format(index_name, unique_field,
                                                                                       doc_id))
                raise ex
            finally:
                if success:
                    writer.commit()
                else:
                    writer.cancel()
        except Exception as ex:
            self.__logger.error('failed to delete document in {0}: {1}'.format(index_name, ex))

        return success

    @replicated
    def put_documents(self, index_name, docs):
        cnt = 0

        try:
            success = False
            writer = self.get_writer(index_name)
            try:
                for doc in docs:
                    self.__logger.debug('indexing document in {0}: {1}'.format(index_name, doc))
                    writer.update_document(**doc)
                    cnt = cnt + 1
                success = True
            except Exception as ex:
                self.__logger.error('failed to index documents in {0} in bulk'.format(index_name))
                raise ex
            finally:
                if success:
                    writer.commit()
                else:
                    cnt = 0  # clear
                    writer.cancel()
        except Exception as ex:
            self.__logger.error('failed to index documents in {0} in bulk: {1}'.format(index_name, ex))

        return cnt

    @replicated
    def delete_documents(self, index_name, doc_ids):
        cnt = 0

        try:
            success = False
            writer = self.get_writer(index_name)
            unique_field = self.get_schema(index_name).get_unique_field()
            try:
                for doc_id in doc_ids:
                    self.__logger.debug('deleting document in {0}: {1}:{2}'.format(index_name, unique_field, doc_id))
                    writer.delete_by_term(unique_field, doc_id)
                    cnt = cnt + 1
                success = True
            except Exception as ex:
                self.__logger.error('failed to delete documents in {0} in bulk'.format(index_name))
                raise ex
            finally:
                if success:
                    writer.commit()
                else:
                    cnt = 0  # clear
                    writer.cancel()
        except Exception as ex:
            self.__logger.error('failed to delete documents in {0} in bulk: {1}'.format(index_name, ex))

        return cnt

    def get_document(self, index_name, doc_id):
        try:
            unique_field = self.get_schema(index_name).get_unique_field()
            try:
                self.__logger.debug('getting document in {0}: {1}:{2}'.format(index_name, unique_field, doc_id))
                doc = self.search_documents(index_name, doc_id, unique_field, 1, page_len=1)
            except Exception as ex:
                self.__logger.error('failed to get document in {0}: {1}:{2}'.format(index_name, unique_field, doc_id))
                raise ex
        except Exception as ex:
            self.__logger.error('failed to get document in {0}'.format(index_name))
            raise ex

        return doc

    def search_documents(self, index_name, query, search_field, page_num, page_len=10, weighting=None, **kwargs):
        try:
            searcher = self.get_searcher(index_name, weighting=weighting)
            query_parser = QueryParser(search_field, self.get_schema(index_name))
            query_obj = query_parser.parse(query)
            try:
                self.__logger.debug('searching documents in {0}: {1}'.format(index_name, query_obj))
                results_page = searcher.search_page(query_obj, page_num, pagelen=page_len, **kwargs)
            except Exception as ex:
                self.__logger.error('failed to search documents in {0}: {1}'.format(index_name, query_obj))
                raise ex
        except Exception as ex:
            self.__logger.error('failed to search documents in {0}'.format(index_name))
            raise ex

        return results_page

    def open_snapshot_file(self):
        try:
            f = open(self.__conf.fullDumpFile, mode='rb')
        except Exception as ex:
            raise ex

        return f



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import unittest
from logging import DEBUG, Formatter, getLogger, INFO, StreamHandler
from tempfile import TemporaryDirectory

import yaml
from pysyncobj import SyncObjConf
from whoosh.filedb.filestore import FileStorage

import cockatrice
from cockatrice.index_server import IndexServer
from cockatrice.schema import Schema


class TestIndexServer(unittest.TestCase):
    def setUp(self):
        self.temp_dir = TemporaryDirectory()
        self.example_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../example'))

        host = '0.0.0.0'
        port = 0
        peer_addrs = []
        dump_file = self.temp_dir.name + '/data.dump'

        index_dir = self.temp_dir.name + '/index'

        logger = getLogger(cockatrice.NAME)
        log_handler = StreamHandler()
        logger.setLevel(DEBUG)
        log_handler.setLevel(INFO)
        log_format = Formatter('%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)d - %(message)s')
        log_handler.setFormatter(log_format)
        logger.addHandler(log_handler)

        conf = SyncObjConf(
            fullDumpFile=dump_file,
            logCompactionMinTime=300,
            dynamicMembershipChange=True
        )

        self.index_server = IndexServer(host, port, peer_addrs, conf, index_dir, logger=logger)

    def tearDown(self):
        self.index_server.stop()
        self.temp_dir.cleanup()

    def test_create_index_from_yaml(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        # create index
        index_name = 'test2_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        # check the number of file file indices
        expected_file_count = 2
        actual_file_count = len(self.index_server.get_file_storage().list())
        self.assertEqual(expected_file_count, actual_file_count)

    def test_create_index_from_json(self):
        with open(self.example_dir + '/schema.json', 'r', encoding='utf-8') as file_obj:
            __dict = json.loads(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        # create index
        index_name = 'test2_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        # check the number of file file indices
        expected_file_count = 2
        actual_file_count = len(self.index_server.get_file_storage().list())
        self.assertEqual(expected_file_count, actual_file_count)

    def test_delete_index(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        # create index
        index_name = 'test2_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        expected_file_count = 2
        actual_file_count = len(self.index_server.get_file_storage().list())
        self.assertEqual(expected_file_count, actual_file_count)

        self.index_server.delete_index(index_name, sync=True)
        expected_file_count = 1
        actual_file_count = len(self.index_server.get_file_storage().list())
        self.assertEqual(expected_file_count, actual_file_count)

    def test_get_index(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        index = self.index_server.create_index(index_name, schema, sync=True)
        self.assertTrue(isinstance(index.storage, FileStorage))

        i = self.index_server.get_index(index_name)
        self.assertTrue(isinstance(i.storage, FileStorage))

    def test_index_document(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        test_doc_id = '1'
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            test_fields = json.loads(file_obj.read(), encoding='utf-8')

        # index document
        self.index_server.put_document(index_name, test_doc_id, test_fields, sync=True)

        # get document
        page = self.index_server.get_document(index_name, test_doc_id)

        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

    def test_delete(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        test_doc_id = '1'
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            test_fields = json.loads(file_obj.read(), encoding='utf-8')

        # index document
        self.index_server.put_document(index_name, test_doc_id, test_fields, sync=True)

        # get document
        page = self.index_server.get_document(index_name, test_doc_id)
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        # delete document
        self.index_server.delete_document(index_name, test_doc_id, sync=True)

        # get document
        page = self.index_server.get_document(index_name, test_doc_id)
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

    def test_index_documents(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        with open(self.example_dir + '/bulk_index.json', 'r', encoding='utf-8') as file_obj:
            test_docs = json.loads(file_obj.read(), encoding='utf-8')

        # index documents in bulk
        self.index_server.put_documents(index_name, test_docs, sync=True)

        page = self.index_server.get_document(index_name, '1')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '2')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '3')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '4')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '5')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

    def test_delete_documents(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        with open(self.example_dir + '/bulk_index.json', 'r', encoding='utf-8') as file_obj:
            test_docs = json.loads(file_obj.read(), encoding='utf-8')

        # index documents in bulk
        self.index_server.put_documents(index_name, test_docs, sync=True)

        page = self.index_server.get_document(index_name, '1')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '2')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '3')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '4')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '5')
        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        with open(self.example_dir + '/bulk_delete.json', 'r', encoding='utf-8') as file_obj:
            test_docs = json.loads(file_obj.read(), encoding='utf-8')

        # index documents in bulk
        self.index_server.delete_documents(index_name, test_docs, sync=True)

        page = self.index_server.get_document(index_name, '1')
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '2')
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '3')
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '4')
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.get_document(index_name, '5')
        expected_count = 0
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

    def test_get_document(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        test_doc_id = '1'
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            test_fields = json.loads(file_obj.read(), encoding='utf-8')

        # index document
        self.index_server.put_document(index_name, test_doc_id, test_fields, sync=True)

        # get document
        page = self.index_server.get_document(index_name, test_doc_id)

        expected_count = 1
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

    def test_search_documents(self):
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())
        schema = Schema(__dict)

        # create file index
        index_name = 'test_file_index'
        self.index_server.create_index(index_name, schema, sync=True)

        with open(self.example_dir + '/bulk_index.json', 'r', encoding='utf-8') as file_obj:
            test_docs = json.loads(file_obj.read(), encoding='utf-8')

        # index documents in bulk
        self.index_server.put_documents(index_name, test_docs, sync=True)

        # search documents
        page = self.index_server.search_documents(index_name, 'search', search_field='text', page_num=1, page_len=10)
        expected_count = 5
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.search_documents(index_name, 'search engine', search_field='text', page_num=1,
                                                  page_len=10)
        expected_count = 3
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.search_documents(index_name, 'distributed search', search_field='text', page_num=1,
                                                  page_len=10)
        expected_count = 2
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)

        page = self.index_server.search_documents(index_name, 'web search', search_field='text', page_num=1,
                                                  page_len=10)
        expected_count = 4
        actual_count = page.total
        self.assertEqual(expected_count, actual_count)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from janome.tokenizer import Tokenizer as Janome
from whoosh.analysis import Token, Tokenizer
from whoosh.compat import text_type


class JanomeTokenizer(Tokenizer):
    def __init__(self, udic='', udic_enc='utf8', udic_type='ipadic', max_unknown_length=1024, wakati=False, mmap=False):
        self.udic = udic
        self.udic_enc = udic_enc
        self.udic_type = udic_type
        self.max_unknown_length = max_unknown_length
        self.wakati = wakati
        self.mmap = mmap

        self.tagger = Janome(udic=self.udic, udic_enc=self.udic_enc, udic_type=self.udic_type,
                             max_unknown_length=self.max_unknown_length, wakati=self.wakati, mmap=self.mmap)

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if k != "tagger"}

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.tagger = Janome(udic=self.udic, udic_enc=self.udic_enc, udic_type=self.udic_type,
                             max_unknown_length=self.max_unknown_length, wakati=self.wakati, mmap=self.mmap)
        self.tagger.tokenize('')

    def __call__(self, value, positions=False, chars=False, keeporiginal=False, removestops=True, start_pos=0,
                 start_char=0, tokenize=True, mode='', **kwargs):
        assert isinstance(value, text_type), '%s is not unicode' % repr(value)

        token = Token(positions, chars, removestops=removestops, mode=mode, **kwargs)

        if not tokenize:
            token.original = token.text = value
            token.boost = 1.0
            if positions:
                token.pos = start_pos
            if chars:
                token.startchar = start_char
                token.endchar = start_char + len(value)
            yield token
        else:
            pos = start_pos
            for janome_token in self.tagger.tokenize(value):
                token.text = janome_token.surface
                token.boost = 1.0
                if keeporiginal:
                    token.original = token.text
                token.stopped = False
                if positions:
                    token.pos = pos
                    pos += 1
                if chars:
                    token.startchar = start_char + janome_token.start
                    token.endchar = token.startchar + len(janome_token.surface)
                yield token



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from cockatrice.scoring import MultiWeighting


class TestMultiWeighting(unittest.TestCase):
    def setUp(self):
        self.example_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../example'))

    def tearDown(self):
        pass

    def test_create_weighting(self):
        weighting_file = self.example_dir + '/weighting.yaml'

        with open(weighting_file, 'r', encoding='utf-8') as file_obj:
            weighting_yaml = file_obj.read()

        weighting = MultiWeighting(weighting_yaml)

        self.assertIsNotNone(weighting)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import unittest
from http import HTTPStatus
from logging import DEBUG, Formatter, getLogger, INFO, StreamHandler
from tempfile import TemporaryDirectory

import yaml
from prometheus_client.core import CollectorRegistry
from pysyncobj import SyncObjConf

from cockatrice import NAME
from cockatrice.index_http_server import IndexHTTPServer
from cockatrice.index_server import IndexServer


class TestIndexHTTPServer(unittest.TestCase):
    def setUp(self):
        self.temp_dir = TemporaryDirectory()
        self.example_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../example'))

        host = '0.0.0.0'
        port = 0
        peer_addrs = []
        dump_file = self.temp_dir.name + '/data.dump'
        http_port = 0

        index_dir = self.temp_dir.name + '/index'

        logger = getLogger(NAME)
        log_handler = StreamHandler()
        logger.setLevel(DEBUG)
        log_handler.setLevel(INFO)
        log_format = Formatter('%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)d - %(message)s')
        log_handler.setFormatter(log_format)
        logger.addHandler(log_handler)

        http_logger = getLogger(NAME + '_http')
        http_log_handler = StreamHandler()
        http_logger.setLevel(INFO)
        http_log_handler.setLevel(INFO)
        http_log_format = Formatter('%(message)s')
        http_log_handler.setFormatter(http_log_format)
        http_logger.addHandler(http_log_handler)

        metrics_registry = CollectorRegistry()

        conf = SyncObjConf(
            fullDumpFile=dump_file,
            logCompactionMinTime=300,
            dynamicMembershipChange=True
        )

        self.index_server = IndexServer(host, port, peer_addrs, conf, index_dir, logger=logger)
        self.index_http_server = IndexHTTPServer(self.index_server, host, http_port, logger=logger,
                                                 http_logger=http_logger, metrics_registry=metrics_registry)

        self.test_client = self.index_http_server.get_test_client()

    def tearDown(self):
        self.index_server.stop()
        self.index_http_server.stop()
        self.temp_dir.cleanup()

    def test_metrics(self):
        response = self.test_client.get('/metrics')
        self.assertEqual(HTTPStatus.OK, response.status_code)

    def test_root(self):
        response = self.test_client.get('/')
        self.assertEqual(HTTPStatus.OK, response.status_code)

    def test_put_index(self):
        index_name = 'test_index'
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/' + index_name, data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

    def test_get_index(self):
        index_name = 'test_index'
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/' + index_name, data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # get index
        response = self.test_client.get('/indices/' + index_name)
        self.assertEqual(HTTPStatus.OK, response.status_code)

    def test_delete_index(self):
        index_name = 'test_index'
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/' + index_name, data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # get index
        response = self.test_client.get('/indices/' + index_name)
        self.assertEqual(HTTPStatus.OK, response.status_code)

        # delete index
        response = self.test_client.delete('/indices/' + index_name, query_string='sync=True')
        self.assertEqual(HTTPStatus.OK, response.status_code)

        # get index
        response = self.test_client.get('/indices/' + index_name)
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_code)

    def test_put_document_yaml(self):
        # read schema
        index_name = 'test_index'
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/' + index_name, data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # read document
        with open(self.example_dir + '/doc1.yaml', 'r', encoding='utf-8') as file_obj:
            doc = file_obj.read()

        # put document
        response = self.test_client.put('/indices/' + index_name + '/documents/1', data=doc,
                                        query_string='sync=True', headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

    def test_put_document_json(self):
        # read schema
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/test_index', data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # read document
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            doc = file_obj.read()

        # put document
        response = self.test_client.put('/indices/test_index/documents/1', data=doc,
                                        query_string='sync=True', headers={'Content-Type': 'application/json'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

    def test_get_document_yaml(self):
        # read schema
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/test_index', data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # read document
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            doc_json = file_obj.read()

        # put document
        response = self.test_client.put('/indices/test_index/documents/1', data=doc_json,
                                        query_string='sync=True', headers={'Content-Type': 'application/json'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # get document
        response = self.test_client.get('/indices/test_index/documents/1?output=yaml')
        self.assertEqual(HTTPStatus.OK, response.status_code)

        data = yaml.safe_load(response.data)
        expected_doc_id = '1'
        actual_doc_id = data['fields']['id']
        self.assertEqual(expected_doc_id, actual_doc_id)

    def test_get_document_json(self):
        # read schema
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/test_index', data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # read document
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            doc_json = file_obj.read()

        # put document
        response = self.test_client.put('/indices/test_index/documents/1', data=doc_json,
                                        query_string='sync=True', headers={'Content-Type': 'application/json'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # get document
        response = self.test_client.get('/indices/test_index/documents/1', query_string='output=json')
        self.assertEqual(HTTPStatus.OK, response.status_code)

        data = json.loads(response.data)
        expected_doc_id = '1'
        actual_doc_id = data['fields']['id']
        self.assertEqual(expected_doc_id, actual_doc_id)

    def test_delete_document(self):
        # read schema
        with open(self.example_dir + '/schema.yaml', 'r', encoding='utf-8') as file_obj:
            schema_yaml = file_obj.read()

        # create index
        response = self.test_client.put('/indices/test_index', data=schema_yaml, query_string='sync=True',
                                        headers={'Content-Type': 'application/yaml'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # read document
        with open(self.example_dir + '/doc1.json', 'r', encoding='utf-8') as file_obj:
            doc_json = file_obj.read()

        # put document
        response = self.test_client.put('/indices/test_index/documents/1', data=doc_json, query_string='sync=True',
                                        headers={'Content-Type': 'application/json'})
        self.assertEqual(HTTPStatus.CREATED, response.status_code)

        # delete document
        response = self.test_client.delete('/indices/test_index/documents/1', query_string='sync=True')
        self.assertEqual(HTTPStatus.OK, response.status_code)

        # get document
        response = self.test_client.get('/indices/test_index/documents/1')
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_code)

    # def test_bulk_index(self):
    #     example_file = self.example_dir + '/bulk_index.json'
    #
    #     file_obj = open(example_file, 'r', encoding='utf-8')
    #     example_data = json.loads(file_obj.read(), encoding='utf-8')
    #     file_obj.close()
    #
    #     sync = True
    #
    #     response = self.client.put('/rest/bulk', data=json.dumps(example_data), query_string='sync=' + str(sync))
    #
    #     expected_status_code = HTTPStatus.CREATED
    #     actual_status_code = response.status_code
    #     self.assertEqual(expected_status_code, actual_status_code)
    #
    #     data = json.loads(response.data)
    #
    #     expected_status_code = HTTPStatus.CREATED
    #     actual_status_code = data['status']['code']
    #     self.assertEqual(expected_status_code, actual_status_code)
    #
    #     expected_count = 5
    #     actual_count = data['count']
    #     self.assertEqual(expected_count, actual_count)
    #
    # def test_bulk_delete(self):
    #     example_bulk_index = self.example_dir + '/bulk_index.json'
    #
    #     file_obj = open(example_bulk_index, 'r', encoding='utf-8')
    #     example_data = json.loads(file_obj.read(), encoding='utf-8')
    #     file_obj.close()
    #
    #     sync = True
    #
    #     response = self.client.put('/rest/bulk', data=json.dumps(example_data), query_string='sync=' + str(sync))
    #
    #     expected_status_code = HTTPStatus.CREATED
    #     actual_status_code = response.status_code
    #     self.assertEqual(expected_status_code, actual_status_code)
    #
    #     example_bulk_delete = self.example_dir + '/bulk_delete.json'
    #
    #     file_obj = open(example_bulk_delete, 'r', encoding='utf-8')
    #     example_data = json.loads(file_obj.read(), encoding='utf-8')
    #     file_obj.close()
    #
    #     sync = True
    #
    #     response = self.client.delete_document('/rest/bulk', data=json.dumps(example_data), query_string='sync=' + str(sync))
    #
    #     expected_status_code = HTTPStatus.OK
    #     actual_status_code = response.status_code
    #     self.assertEqual(expected_status_code, actual_status_code)
    #
    #     data = json.loads(response.data)
    #
    #     expected_status_code = HTTPStatus.OK
    #     actual_status_code = data['status']['code']
    #     self.assertEqual(expected_status_code, actual_status_code)
    #
    #     expected_count = 5
    #     actual_count = data['count']
    #     self.assertEqual(expected_count, actual_count)
    #
    # def test_search(self):
    #     pass



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib


def get_instance(class_name, **class_args):
    class_data = class_name.split('.')

    module_path = '.'.join(class_data[:-1])
    class_name = class_data[-1]

    module = importlib.import_module(module_path)
    class_obj = getattr(module, class_name)

    if class_args:
        return class_obj(**class_args)
    else:
        return class_obj()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket


def get_ipv4(host):
    ipv4 = None

    try:
        addrlist = socket.getaddrinfo(host, None, 0, 0, 0, socket.AI_CANONNAME)
        for family, kind, proto, canonical, sockaddr in addrlist:
            if family != socket.AF_INET:
                continue
            if kind == socket.SOCK_STREAM:
                if canonical == host or canonical == '':
                    ipv4 = sockaddr[0]
                else:
                    ipv4 = get_ipv4(canonical)
            if ipv4 is not None:
                break
    except Exception as ex:
        raise ex

    return ipv4



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pysyncobj.encryptor import getEncryptor
from pysyncobj.poller import createPoller
from pysyncobj.tcp_connection import TcpConnection

from cockatrice.util.resolver import get_ipv4


class Executor:
    def __init__(self, cmd, args=None, bind_addr='localhost:7070', password=None, timeout=1):
        try:
            self.__result = None

            self.__request = [cmd]
            if args is not None:
                self.__request.extend(args)

            host, port = bind_addr.split(":", 1)

            self.__host = get_ipv4(host)
            self.__port = int(port)
            self.__password = password
            self.__poller = createPoller('auto')
            self.__connection = TcpConnection(self.__poller, onMessageReceived=self.__on_message_received,
                                              onConnected=self.__on_connected, onDisconnected=self.__on_disconnected,
                                              socket=None, timeout=10.0, sendBufferSize=2 ** 13, recvBufferSize=2 ** 13)
            if self.__password is not None:
                self.__connection.encryptor = getEncryptor(self.__password)
            self.__is_connected = self.__connection.connect(self.__host, self.__port)
            while self.__is_connected:
                self.__poller.poll(timeout)
        except Exception as ex:
            raise ex

    def __on_message_received(self, message):
        if self.__connection.encryptor and not self.__connection.sendRandKey:
            self.__connection.sendRandKey = message
            self.__connection.send(self.__request)
            return

        self.__result = message
        self.__connection.disconnect()

    def __on_connected(self):
        if self.__connection.encryptor:
            self.__connection.recvRandKey = os.urandom(32)
            self.__connection.send(self.__connection.recvRandKey)
            return

        self.__connection.send(self.__request)

    def __on_disconnected(self):
        self.__is_connected = False

    def get_result(self):
        return self.__result


def execute(cmd, args=None, bind_addr='127.0.0.1:7070', password=None, timeout=1):
    try:
        response = Executor(cmd, args=args, bind_addr=bind_addr, password=password, timeout=timeout).get_result()
    except Exception as ex:
        raise ex

    return response


def get_status(bind_addr='127.0.0.1:7070', password=None, timeout=1):
    return execute('status', args=None, bind_addr=bind_addr, password=password, timeout=timeout)


def add_node(node_name, bind_addr='127.0.0.1:7070', password=None, timeout=1):
    return execute('add', args=[node_name], bind_addr=bind_addr, password=password, timeout=timeout)


def delete_node(node_name, bind_addr='127.0.0.1:7070', password=None, timeout=1):
    return execute('remove', args=[node_name], bind_addr=bind_addr, password=password, timeout=timeout)


def get_snapshot(bind_addr='127.0.0.1:7070', password=None, timeout=1):
    return execute('get_snapshot', args=None, bind_addr=bind_addr, password=password, timeout=timeout)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
from http import HTTPStatus
from logging import getLogger
from threading import Thread

import yaml
from flask import after_this_request, Flask, request, Response
from prometheus_client.core import CollectorRegistry, Counter, Gauge, Histogram
from prometheus_client.exposition import CONTENT_TYPE_LATEST, generate_latest
from werkzeug.serving import make_server
from whoosh.scoring import BM25F

from cockatrice import NAME, VERSION
from cockatrice.schema import Schema
from cockatrice.scoring import get_multi_weighting

TRUE_STRINGS = ['true', 'yes', 'on', 't', 'y', '1']


class ServerThread(Thread):
    def __init__(self, host, port, app, logger=getLogger()):
        self.__logger = logger
        self.__logger.info('creating server thread')

        Thread.__init__(self)
        self.srv = make_server(host, port, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.__logger.info('starting server thread')
        self.srv.serve_forever()

    def shutdown(self):
        self.__logger.info('stopping server thread')
        self.srv.shutdown()


class IndexHTTPServer:
    def __init__(self, index_server, host='localhost', port=8080, logger=getLogger(), http_logger=getLogger(),
                 metrics_registry=CollectorRegistry()):
        self.__logger = logger
        self.__http_logger = http_logger
        self.__metrics_registry = metrics_registry

        self.__host = host
        self.__port = port
        self.__index_server = index_server

        self.__logger.info('starting index http server: {0}'.format(self.__port))

        self.__app = Flask('index_http_server')
        self.__app.add_url_rule('/', endpoint='root', view_func=self.__root, methods=['GET'])
        self.__app.add_url_rule('/indices/<index_name>', endpoint='get_index', view_func=self.__get_index,
                                methods=['GET'])
        self.__app.add_url_rule('/indices/<index_name>', endpoint='create_index', view_func=self.__create_index,
                                methods=['PUT'])
        self.__app.add_url_rule('/indices/<index_name>', endpoint='delete_index', view_func=self.__delete_index,
                                methods=['DELETE'])
        self.__app.add_url_rule('/indices/<index_name>/documents/<doc_id>', endpoint='get_document',
                                view_func=self.__get_document, methods=['GET'])
        self.__app.add_url_rule('/indices/<index_name>/documents/<doc_id>', endpoint='put_document',
                                view_func=self.__put_document, methods=['PUT'])
        self.__app.add_url_rule('/indices/<index_name>/documents/<doc_id>', endpoint='delete_document',
                                view_func=self.__delete_document, methods=['DELETE'])
        self.__app.add_url_rule('/indices/<index_name>/documents', endpoint='put_documents',
                                view_func=self.__put_documents, methods=['PUT'])
        self.__app.add_url_rule('/indices/<index_name>/documents', endpoint='delete_documents',
                                view_func=self.__delete_documents, methods=['DELETE'])
        self.__app.add_url_rule('/indices/<index_name>/search', endpoint='search_documents',
                                view_func=self.__search_documents, methods=['GET', 'POST'])
        self.__app.add_url_rule('/indices/<index_name>/optimize', endpoint='optimize_index',
                                view_func=self.__optimize_index, methods=['GET'])
        self.__app.add_url_rule('/cluster', endpoint='get_cluster', view_func=self.__get_cluster, methods=['GET'])
        self.__app.add_url_rule('/cluster/<node_name>', endpoint='put_node', view_func=self.__put_node, methods=['PUT'])
        self.__app.add_url_rule('/cluster/<node_name>', endpoint='delete_node', view_func=self.__delete_node,
                                methods=['DELETE'])
        self.__app.add_url_rule('/metrics', endpoint='metrics', view_func=self.__metrics, methods=['GET'])
        self.__app.add_url_rule('/health/liveness', endpoint='liveness', view_func=self.__liveness, methods=['GET'])
        self.__app.add_url_rule('/health/readiness', endpoint='readiness', view_func=self.__readiness, methods=['GET'])
        self.__app.add_url_rule('/snapshot', endpoint='get_snapshot', view_func=self.__get_snapshot, methods=['GET'])
        self.__app.add_url_rule('/snapshot', endpoint='put_snapshot', view_func=self.__put_snapshot, methods=['PUT'])

        # disable Flask default logger
        self.__app.logger.disabled = True
        getLogger('werkzeug').disabled = True

        # metrics
        self.__metrics_http_requests_total = Counter(
            '{0}_http_requests_total'.format(NAME),
            'The number of requests.',
            [
                'method',
                'endpoint',
                'status_code'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_requests_bytes_total = Counter(
            '{0}_http_requests_bytes_total'.format(NAME),
            'A summary of the invocation requests bytes.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_responses_bytes_total = Counter(
            '{0}_http_responses_bytes_total'.format(NAME),
            'A summary of the invocation responses bytes.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_http_requests_duration_seconds = Histogram(
            '{0}_http_requests_duration_seconds'.format(NAME),
            'The invocation duration in seconds.',
            [
                'method',
                'endpoint'
            ],
            registry=self.__metrics_registry
        )
        self.__metrics_index_documents = Gauge(
            '{0}_index_documents'.format(NAME),
            'The number of documents.',
            [
                'index_name',
            ],
            registry=self.__metrics_registry
        )

        self.__server_thread = None
        try:
            # run server
            self.__server_thread = ServerThread(self.__host, self.__port, self.__app, logger=self.__logger)
            self.__server_thread.start()
            self.__logger.info('index http server started')
        except Exception as ex:
            self.__logger.critical(ex)

    def stop(self):
        self.__logger.info('stopping index http server')
        self.__server_thread.shutdown()

    def __record_http_log(self, req, resp):
        log_message = '{0} - {1} [{2}] "{3} {4} {5}" {6} {7} "{8}" "{9}"'.format(
            req.remote_addr,
            req.remote_user if req.remote_user is not None else '-',
            time.strftime('%d/%b/%Y %H:%M:%S +0000', time.gmtime()),
            req.method,
            req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else ''),
            req.environ.get('SERVER_PROTOCOL'),
            resp.status_code,
            resp.content_length,
            req.referrer if req.referrer is not None else '-',
            req.user_agent
        )
        self.__http_logger.info(log_message)

        return

    def __record_http_metrics(self, start_time, req, resp):
        self.__metrics_http_requests_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else ''),
            status_code=resp.status_code.value
        ).inc()

        self.__metrics_http_requests_bytes_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).inc(req.content_length if req.content_length is not None else 0)

        self.__metrics_http_responses_bytes_total.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).inc(resp.content_length if resp.content_length is not None else 0)

        self.__metrics_http_requests_duration_seconds.labels(
            method=req.method,
            endpoint=req.path + ('?{0}'.format(req.query_string.decode('utf-8')) if len(req.query_string) > 0 else '')
        ).observe(time.time() - start_time)

        return

    def __record_index_metrics(self, index_name):
        doc_count = self.__index_server.get_doc_count(index_name)
        if doc_count is not None:
            self.__metrics_index_documents.labels(
                index_name=index_name,
            ).set(doc_count)
        else:
            self.__metrics_index_documents.labels(
                index_name=index_name,
            ).set(0)

        return

    def __root(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        resp = Response()
        try:
            resp.status_code = HTTPStatus.OK
            resp.content_type = 'text/plain; charset="UTF-8"'
            resp.data = NAME + ' ' + VERSION + ' is running.\n'
        except Exception as ex:
            resp.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            resp.content_type = 'text/plain; charset="UTF-8"'
            resp.data = '{0}\n{1}'.format(resp.status_code.phrase, resp.status_code.description)
            self.__logger.error(ex)

        return resp

    def __get_index(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            index = self.__index_server.get_index(index_name)
            if index is None:
                raise KeyError('index does not exist')

            data['index'] = {}
            data['index']['name'] = index_name
            data['index']['doc_count'] = index.doc_count()
            data['index']['doc_count_all'] = index.doc_count_all()
            data['index']['last_modified'] = index.last_modified()
            data['index']['latest_generation'] = index.latest_generation()
            data['index']['version'] = index.version
            data['index']['storage'] = {
                'folder': index.storage.folder,
                'supports_mmap': index.storage.supports_mmap,
                'readonly': index.storage.readonly,
                'files': index.storage.list()
            }
            status_code = HTTPStatus.OK
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __create_index(self, index_name):
        start_time = time.time()

        data = {}
        status_code = None

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            try:
                if 'application/yaml' in request.headers.get('Content-Type'):
                    schema = Schema(yaml.safe_load(request.data.decode("utf-8")))
                elif 'application/json' in request.headers.get('Content-Type'):
                    schema = Schema(json.loads(request.data.decode("utf-8")))
                else:
                    raise ValueError('unsupported Content-Type')
            except Exception as ex:
                raise ex

            index = self.__index_server.create_index(index_name, schema, sync=sync)

            if sync:
                if index is None:
                    raise ValueError('failed to create index')
                else:
                    status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except (json.decoder.JSONDecodeError, ValueError) as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __delete_index(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__index_server.delete_index(index_name, sync=sync)

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __optimize_index(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__index_server.optimize_index(index_name, sync=sync)

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __get_document(self, index_name, doc_id):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            results_page = self.__index_server.get_document(index_name, doc_id)

            if results_page.total <= 0:
                raise KeyError('{0} does not exist in {1}'.format(doc_id, index_name))

            hit = results_page.results[0]
            fields = {}
            for i in hit.iteritems():
                fields[i[0]] = i[1]
            data['fields'] = fields

            status_code = HTTPStatus.OK
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __put_document(self, index_name, doc_id):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            try:
                if 'application/yaml' in request.headers.get('Content-Type'):
                    print(request.get_data(as_text=False, parse_form_data=True))
                    doc = yaml.safe_load(request.data.decode("utf-8"))
                elif 'application/json' in request.headers.get('Content-Type'):
                    doc = json.loads(request.data.decode("utf-8"))
                else:
                    raise ValueError('unsupported Content-Type')
            except Exception as ex:
                raise ex

            self.__index_server.put_document(index_name, doc_id, doc, sync=sync)

            if sync:
                status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except json.decoder.JSONDecodeError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __delete_document(self, index_name, doc_id):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            self.__index_server.delete_document(index_name, doc_id, sync=sync)

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __put_documents(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            try:
                if 'application/yaml' in request.headers.get('Content-Type'):
                    docs = yaml.safe_load(request.data)
                elif 'application/json' in request.headers.get('Content-Type'):
                    docs = json.loads(request.data)
                else:
                    raise ValueError('unsupported Content-Type')
            except Exception as ex:
                raise ex

            cnt = self.__index_server.put_documents(index_name, docs, sync=sync)
            if cnt is not None:
                data['count'] = cnt

            if sync:
                status_code = HTTPStatus.CREATED
            else:
                status_code = HTTPStatus.ACCEPTED
        except json.decoder.JSONDecodeError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.BAD_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __delete_documents(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            sync = False
            if request.args.get('sync', default='', type=str).lower() in TRUE_STRINGS:
                sync = True

            try:
                if 'application/yaml' in request.headers.get('Content-Type'):
                    doc_ids = yaml.safe_load(request.data)
                elif 'application/json' in request.headers.get('Content-Type'):
                    doc_ids = json.loads(request.data)
                else:
                    raise ValueError('unsupported Content-Type')
            except Exception as ex:
                raise ex

            cnt = self.__index_server.delete_documents(index_name, doc_ids, sync=sync)
            if cnt is not None:
                data['count'] = cnt

            if sync:
                status_code = HTTPStatus.OK
            else:
                status_code = HTTPStatus.ACCEPTED
        except KeyError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.NOT_FOUND
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __search_documents(self, index_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            query = request.args.get('query', default='', type=str)
            search_field = request.args.get('search_field', default=self.__index_server.get_index(
                index_name).schema.get_default_search_field(), type=str)
            page_num = request.args.get('page_num', default=1, type=int)
            page_len = request.args.get('page_len', default=10, type=int)
            weighting = BM25F
            if len(request.data) > 0:
                try:
                    if 'application/yaml' in request.headers.get('Content-Type'):
                        weighting = get_multi_weighting(yaml.safe_load(request.data))
                    elif 'application/json' in request.headers.get('Content-Type'):
                        weighting = get_multi_weighting(json.loads(request.data))
                    else:
                        raise ValueError('unsupported Content-Type')
                except Exception as ex:
                    self.__logger.error('failed to load weighting config: {0}'.format(ex))

            results_page = self.__index_server.search_documents(index_name, query, search_field, page_num,
                                                                page_len=page_len, weighting=weighting)

            data['results'] = {
                'is_last_page': results_page.is_last_page(),
                'page_count': results_page.pagecount,
                'page_len': results_page.pagelen,
                'page_num': results_page.pagenum,
                'total': results_page.total
            }
            if results_page.pagecount >= page_num or results_page.total <= 0:
                hits = []
                for result in results_page.results[results_page.offset:]:
                    doc = {'fields': {}}
                    for item in result.iteritems():
                        doc['fields'][item[0]] = item[1]
                    hit = {
                        'doc': doc,
                        'score': result.score,
                        'rank': result.rank,
                        'pos': result.pos
                    }
                    hits.append(hit)
                data['results']['hits'] = hits
            else:
                raise ValueError('page_num must be <= {0}'.format(results_page.pagecount))

            status_code = HTTPStatus.OK
        except ValueError as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.MISDIRECTED_REQUEST
            self.__logger.error(ex)
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __get_node(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            data['node'] = self.__index_server.getStatus()
            status_code = HTTPStatus.OK
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __get_cluster(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            data['cluster'] = self.__index_server.getStatus()
            status_code = HTTPStatus.OK
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __put_node(self, node_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            self.__index_server.addNodeToCluster(node_name)

            status_code = HTTPStatus.OK
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __delete_node(self, node_name):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            self.__index_server.removeNodeFromCluster(node_name)

            data['cluster'] = self.__index_server.getStatus()

            status_code = HTTPStatus.OK
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __metrics(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        # index metrics
        for index_name in self.__index_server.get_indices().keys():
            self.__record_index_metrics(index_name)

        resp = Response()
        try:
            resp.status_code = HTTPStatus.OK
            resp.content_type = CONTENT_TYPE_LATEST
            resp.data = generate_latest(self.__metrics_registry)
        except Exception as ex:
            resp.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            resp.content_type = 'text/plain; charset="UTF-8"'
            resp.data = '{0}\n{1}'.format(resp.status_code.phrase, resp.status_code.description)
            self.__logger.error(ex)

        return resp

    def __liveness(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            data['liveness'] = True
            status_code = HTTPStatus.OK
        except Exception as ex:
            data['liveness'] = False
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __readiness(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        data = {}
        status_code = None
        try:
            if self.__index_server.isReady():
                data['readiness'] = True
                status_code = HTTPStatus.OK
            else:
                raise Exception('index server is not ready')
        except Exception as ex:
            data['readiness'] = False
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def __get_snapshot(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, response)
            self.__record_http_metrics(start_time, request, response)
            return response

        resp = Response()
        try:
            resp.status_code = HTTPStatus.OK
            resp.headers['Content-Type'] = 'application/zip'
            with self.__index_server.open_snapshot_file() as f:
                resp.headers['Content-Disposition'] = 'attachment; filename=' + f.name
                resp.data = f.read()
        except FileNotFoundError as ex:
            resp.status_code = HTTPStatus.NOT_FOUND
            resp.content_type = 'text/plain; charset="UTF-8"'
            resp.data = '{0}\n{1}'.format(resp.status_code.phrase, resp.status_code.description)
            self.__logger.error(ex)
        except Exception as ex:
            resp.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            resp.content_type = 'text/plain; charset="UTF-8"'
            resp.data = '{0}\n{1}'.format(resp.status_code.phrase, resp.status_code.description)
            self.__logger.error(ex)

        return resp

    def __put_snapshot(self):
        start_time = time.time()

        @after_this_request
        def to_do_after_this_request(response):
            self.__record_http_log(request, resp)
            self.__record_http_metrics(start_time, request, resp)
            return response

        data = {}
        status_code = None
        try:
            self.__index_server.forceLogCompaction()

            status_code = HTTPStatus.ACCEPTED
        except Exception as ex:
            data['error'] = '{0}'.format(ex.args[0])
            status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            self.__logger.error(ex)
        finally:
            data['time'] = time.time() - start_time
            data['status'] = {'code': status_code.value, 'phrase': status_code.phrase,
                              'description': status_code.description}

        output = request.args.get('output', default='json', type=str).lower()

        # make response
        resp = make_response(data, output)
        resp.status_code = status_code

        return resp

    def get_test_client(self, ues_cookies=True, **kwargs):
        return self.__app.test_client(ues_cookies, **kwargs)


def make_response(data, output='json'):
    resp = Response()

    if output == 'json':
        resp.data = json.dumps(data, indent=2)
        resp.content_type = 'application/json; charset="UTF-8"'
    elif output == 'yaml':
        resp.data = yaml.safe_dump(data, default_flow_style=False, indent=2)
        resp.content_type = 'application/yaml; charset="UTF-8"'

    return resp



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import getLogger
from os import path
from threading import Thread

from prometheus_client.core import CollectorRegistry
from pysyncobj import SyncObjConf

from cockatrice.command import add_node, get_snapshot, get_status
from cockatrice.index_http_server import IndexHTTPServer
from cockatrice.index_server import IndexServer


class IndexNode:
    def __init__(self, host='localhost', port=7070, seed_addr=None, conf=SyncObjConf(),
                 index_dir='/tmp/cockatrice/index', http_port=8080, logger=getLogger(), http_logger=getLogger(),
                 metrics_registry=CollectorRegistry()):
        self.__host = host
        self.__port = port
        self.__seed_addr = seed_addr
        self.__peer_addrs = []
        self.__conf = conf
        self.__index_dir = index_dir
        self.__http_port = http_port
        self.__logger = logger
        self.__http_logger = http_logger
        self.__metrics_registry = metrics_registry

        self.__logger.info('starting index node')

        if self.__seed_addr is not None:
            bind_addr = '{0}:{1}'.format(self.__host, self.__port)

            # execute a command to get status from the cluster
            self.__logger.info('getting cluster status via {0}'.format(self.__seed_addr))
            status_result = get_status(bind_addr=self.__seed_addr, timeout=0.5)
            if status_result is not None:
                # get peer addresses from above command result
                self_addr = status_result['self']
                if self_addr not in self.__peer_addrs:
                    self.__peer_addrs.append(self_addr)
                for k in status_result.keys():
                    if k.startswith('partner_node_status_server_'):
                        partner_addr = k[len('partner_node_status_server_'):]
                        if partner_addr not in self.__peer_addrs:
                            self.__peer_addrs.append(partner_addr)

                # add this node to the cluster
                if bind_addr not in self.__peer_addrs:
                    self.__logger.info('request to add {0} to {1}'.format(bind_addr, self.__seed_addr))
                    Thread(target=add_node,
                           kwargs={'node_name': bind_addr, 'bind_addr': self.__seed_addr, 'timeout': 0.5}).start()

                # remove this node's address from peer addresses
                if bind_addr in self.__peer_addrs:
                    self.__peer_addrs.remove(bind_addr)

                # get leader node
                leader = status_result['leader']

                # copy snapshot from the leader node
                if not path.exists(self.__conf.fullDumpFile):
                    try:
                        self.__logger.info('copying snapshot from {0}'.format(leader))
                        snapshot = get_snapshot(bind_addr=leader, timeout=0.5)
                        if snapshot is not None:
                            with open(self.__conf.fullDumpFile, 'wb') as f:
                                f.write(snapshot)
                            self.__logger.info('snapshot copied from {0}'.format(leader))
                    except Exception as ex:
                        self.__logger.error('failed to copy snapshot from {0}: {1}'.format(leader, ex))
            else:
                self.__logger.error('failed to get cluster status via {0}'.format(self.__seed_addr))

        self.__index_server = IndexServer(host=self.__host, port=self.__port, peer_addrs=self.__peer_addrs,
                                          conf=self.__conf, index_dir=self.__index_dir, logger=self.__logger)
        self.__index_http_server = IndexHTTPServer(self.__index_server, host=self.__host, port=self.__http_port,
                                                   logger=self.__logger, http_logger=self.__http_logger,
                                                   metrics_registry=self.__metrics_registry)

        self.__logger.info('index node started')

    def stop(self):
        self.__logger.info('stopping index node')
        self.__index_http_server.stop()
        self.__index_server.stop()
        self.__logger.info('index node stopped')



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import unittest
from tempfile import TemporaryDirectory

import yaml
from whoosh.index import create_in

from cockatrice.schema import Schema


class TestSchema(unittest.TestCase):
    def setUp(self):
        self.temp_dir = TemporaryDirectory()
        self.index_dir = self.temp_dir.name + '/index'
        self.example_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../example'))

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_yaml(self):
        schema_file = self.example_dir + '/schema.yaml'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())

        schema = Schema(__dict)

        self.assertIsNotNone(schema)

    def test_json(self):
        schema_file = self.example_dir + '/schema.json'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = json.loads(file_obj.read())

        schema = Schema(__dict)

        self.assertIsNotNone(schema)

    def test_yaml_create_index(self):
        schema_file = self.example_dir + '/schema.yaml'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())

        schema = Schema(__dict)

        if self.index_dir is not None:
            os.makedirs(self.index_dir, exist_ok=True)

        ix = create_in(self.index_dir, schema)

        self.assertIsNotNone(ix)

    def test_json_create_index(self):
        schema_file = self.example_dir + '/schema.json'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = json.loads(file_obj.read())

        schema = Schema(__dict)

        if self.index_dir is not None:
            os.makedirs(self.index_dir, exist_ok=True)

        ix = create_in(self.index_dir, schema)

        self.assertIsNotNone(ix)

    def test_yaml_get_unique_field(self):
        schema_file = self.example_dir + '/schema.yaml'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())

        schema = Schema(__dict)

        expected = 'id'
        actual = schema.get_unique_field()

        self.assertEqual(expected, actual)

    def test_json_get_unique_field(self):
        schema_file = self.example_dir + '/schema.json'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = json.loads(file_obj.read())

        schema = Schema(__dict)

        expected = 'id'
        actual = schema.get_unique_field()

        self.assertEqual(expected, actual)

    def test_yaml_get_default_search_field(self):
        schema_file = self.example_dir + '/schema.yaml'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = yaml.safe_load(file_obj.read())

        schema = Schema(__dict)

        expected = 'text'
        actual = schema.get_default_search_field()

        self.assertEqual(expected, actual)

    def test_json_get_default_search_field(self):
        schema_file = self.example_dir + '/schema.json'
        with open(schema_file, 'r', encoding='utf-8') as file_obj:
            __dict = json.loads(file_obj.read())

        schema = Schema(__dict)

        expected = 'text'
        actual = schema.get_default_search_field()

        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from whoosh.scoring import BM25F, MultiWeighting

from cockatrice.util.loader import get_instance


def get_weighting(class_name, **class_args):
    try:
        weighting = get_instance(class_name, **class_args)
    except Exception as ex:
        raise ex

    return weighting


def get_multi_weighting(weighting_dict):
    try:
        default_weighting = BM25F
        field_weighting = {}

        for field_name in weighting_dict['weighting'].keys():
            class_name = weighting_dict['weighting'][field_name]['class']
            class_args = weighting_dict['weighting'][field_name]['args'] if 'args' in weighting_dict['weighting'][
                field_name] else {}
            instance = get_instance(class_name, **class_args)
            if field_name == 'default':
                default_weighting = instance
            else:
                field_weighting[field_name] = instance

        weighting = MultiWeighting(default_weighting, **field_weighting)
    except Exception as ex:
        raise ex

    return weighting
